# pylint: disable = missing-module-docstring, missing-class-docstring
from reporter.base import RestManager
from reporter.mixins import CreateMixin
from reporter.objects.finding_comment import FindingComment

__all__ = [
    "FindingEventReplyManager",
]


class FindingEventReplyManager(RestManager, CreateMixin):
    _path = "finding-events/{finding_event_id}/finding-comments"
    _parent_attrs = {"finding_event_id": "id"}
    _obj_cls = FindingComment



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, GetMixin, ListMixin, UpdateMixin

__all__ = [
    "User",
    "UserManager",
]


class User(RestObject):
    pass


class UserManager(RestManager, CreateMixin, GetMixin, ListMixin, UpdateMixin):
    _path = "users"
    _obj_cls = User



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin


__all__ = [
    "OutputFile",
    "OutputFileManager",
    "AssessmentOutputFileManager",
]


class OutputFile(RestObject):
    pass


class OutputFileManager(RestManager, DeleteMixin):
    _path = "output-files"
    _obj_cls = OutputFile


class AssessmentOutputFileManager(RestManager, CreateMixin):
    _path = "assessments/{assessment_id}/output-files"
    _parent_attrs = {"assessment_id": "id"}
    _obj_cls = OutputFile



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import UpdateMixin

__all__ = [
    "AssessmentPhase",
    "AssessmentPhaseManager",
]


class AssessmentPhase(RestObject):
    pass


class AssessmentPhaseManager(RestManager, UpdateMixin):
    _path = "assessment-phases"
    _obj_cls = AssessmentPhase



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, GetRawMixin

__all__ = [
    "Document",
    "DocumentManager",
]


class Document(RestObject):
    pass


class DocumentManager(RestManager, DeleteMixin, CreateMixin, GetRawMixin):
    _path = "documents"
    _obj_cls = Document



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, UpdateMixin

__all__ = [
    "AssessmentUser",
    "AssessmentAssessmentUserManager",
]


class AssessmentUser(RestObject):
    pass


class AssessmentAssessmentUserManager(
    RestManager,
    CreateMixin,
    UpdateMixin,
    DeleteMixin,
):
    _path = "assessments/{assessment_id}/users"
    _parent_attrs = {"assessment_id": "id"}
    _obj_cls = AssessmentUser



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import GetMixin, UpdateMixin

__all__ = [
    "AssessmentSection",
    "AssessmentSectionManager",
]


class AssessmentSection(RestObject):
    pass


class AssessmentSectionManager(RestManager, GetMixin, UpdateMixin):
    _path = "assessment-sections"
    _obj_cls = AssessmentSection



# pylint: disable = missing-module-docstring, missing-class-docstring, redefined-builtin
from typing import Any

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, GetMixin, ListMixin, UpdateMixin, DeleteMixin


__all__ = [
    "Assessment",
    "AssessmentManager",
    "ClientAssessmentManager",
]


class Assessment(RestObject):
    pass


class AssessmentManager(RestManager, GetMixin, ListMixin, UpdateMixin, DeleteMixin):
    _path = "assessments"
    _obj_cls = Assessment

    def get_full_report(self, id, **kwargs: Any) -> bytes:
        """Get the full PDF report of an assessment.

        Args:
            id: The ID of the assessment
            **kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The full report as a bytestring.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """
        path = f"{self._path}/{id}/pdf-reports/full"

        result = self.reporter.http_request(
            verb="get",
            path=path,
            headers={"Accept": "*/*"},
            **kwargs,
        )

        return result.content

    def get_management_report(self, id, **kwargs: Any) -> bytes:
        """Get the management PDF report of an assessment.

        Args:
            id: The ID of the assessment
            **kwargs: Extra options to pass to the underlying
                :func:`reporter.Reporter.http_request` call.

        Returns:
            The management report as a bytestring.

        Raises:
            ReporterHttpError: If raised by the underlying call to
                :func:`reporter.Reporter.http_request`.

        """
        path = f"{self._path}/{id}/pdf-reports/management"

        return self.reporter.get_raw_file(path, **kwargs)


class ClientAssessmentManager(RestManager, CreateMixin):
    _path = "clients/{client_id}/assessments"
    _parent_attrs = {"client_id": "id"}
    _obj_cls = Assessment



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, UpdateMixin

__all__ = [
    "FindingRetest",
    "FindingRetestManager",
    "FindingFindingRetestManager",
]


class FindingRetest(RestObject):
    pass


class FindingRetestManager(RestManager, UpdateMixin, DeleteMixin):
    _path = "finding-retests"
    _obj_cls = FindingRetest


class FindingFindingRetestManager(RestManager, CreateMixin):
    _path = "findings/{finding_id}/finding-retests"
    _parent_attrs = {"finding_id": "id"}
    _obj_cls = FindingRetest



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CreateMixin, DeleteMixin, GetMixin, ListMixin, UpdateMixin

__all__ = [
    "UserGroup",
    "UserGroupManager",
    "ClientUserGroupManager",
]


class UserGroup(RestObject):
    pass


class UserGroupManager(
    RestManager,
    DeleteMixin,
    GetMixin,
    ListMixin,
    UpdateMixin,
):
    _path = "user-groups"
    _obj_cls = UserGroup


class ClientUserGroupManager(RestManager, CreateMixin):
    _path = "clients/{client_id}/user-groups"
    _parent_attrs = {"client_id": "id"}
    _obj_cls = UserGroup



"""This module contains exceptions raised by this library."""

from typing import Optional, Union

__all__ = [
    "ReporterError",
    "ReporterHttpError",
]


class ReporterError(Exception):
    """Base class for Reporter errors.

    Args:
       error_message: The error message.
       response_code: The response code returned by the server.
       response_body: The response body returned by the server.

    """

    def __init__(
        self,
        error_message: Union[bytes, str] = "",
        response_code: Optional[int] = None,
        response_body: Optional[bytes] = None,
    ) -> None:
        Exception.__init__(self, error_message)

        if isinstance(error_message, bytes):
            self.error_message = error_message.decode()
        else:
            self.error_message = error_message

        self.response_code = response_code
        self.response_body = response_body

    def __str__(self) -> str:
        if self.response_code is not None:
            return f"{self.response_code}: {self.error_message}"
        return f"{self.error_message}"


class ReporterHttpError(ReporterError):
    """Raised on unsuccessful HTTP response."""



# pylint: disable = missing-module-docstring, missing-class-docstring
from reporter.base import RestManager
from reporter.mixins import CreateMixin
from reporter.objects.assessment_section_comment import AssessmentSectionComment

__all__ = [
    "AssessmentSectionEventReplyManager",
]


class AssessmentSectionEventReplyManager(RestManager, CreateMixin):
    _path = "assessment-section-events/{assessment_section_event_id}/assessment-section-comments"
    _parent_attrs = {"assessment_section_event_id": "id"}
    _obj_cls = AssessmentSectionComment



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import (
    CrudMixin,
    ListMixin,
    SearchMixin,
)

__all__ = [
    "FindingTemplate",
    "FindingTemplateManager",
]


class FindingTemplate(RestObject):
    pass


class FindingTemplateManager(
    RestManager,
    CrudMixin,
    ListMixin,
    SearchMixin,
):
    _path = "finding-templates"
    _obj_cls = FindingTemplate



"""Wrapper around the Reporter API."""

__version__ = "0.1.0"

from reporter.base import *
from reporter.client import *
from reporter.exceptions import *
from reporter.objects import *
from reporter.object_children import *
from reporter.object_includes import *



"""Base classes for API models.

This module contains base classes that should be inherited by objects
representing API models, managers of these objects, and lists of these objects.

"""

from collections.abc import Sequence
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
)

from reporter.client import Reporter

__all__ = [
    "RestObject",
    "RestList",
    "RestManager",
]


class RestObject:
    """Represents an object built from server data.

    Args:
        reporter: The :class:`~reporter.Reporter` instance used by this object to
            perform requests.
        attrs: Object attributes
    """

    reporter: Reporter

    _attrs: Dict[str, Any]
    _includes: Mapping[str, Type["RestObject"]] = {}
    _children: Mapping[str, Type["RestManager"]] = {}

    def __init__(self, reporter: Reporter, attrs: Mapping[str, Any]) -> None:
        self.reporter = reporter
        self._attrs = dict(attrs)
        for key, mgr_cls in self._children.items():
            setattr(self, key, mgr_cls(self.reporter, parent=self))
        self._deserialize_includes()

    def __getattr__(self, attr: str) -> Any:
        try:
            return self._attrs[attr]
        except KeyError as exc:
            message = f"{type(self).__name__!r} object has no attribute '{attr}'"
            raise AttributeError(message) from exc

    def __getitem__(self, attr: str) -> Any:
        val = self._attrs[attr]
        if isinstance(val, RestObject):
            return dict(self._attrs[attr])
        if isinstance(val, list) and len(val) > 0 and isinstance(val[0], RestObject):
            # We assume that either all elements of val are RestObjects or none
            # of them are.
            return [dict(v) for v in val]
        return val

    def __iter__(self) -> Any:
        return iter(self._attrs)

    def __len__(self) -> int:
        return len(self._attrs)

    def __dir__(self) -> Iterable[str]:
        return set(self._attrs.keys()).union(super().__dir__())

    def __eq__(self, other: object):
        if not isinstance(other, RestObject):
            return NotImplemented
        return self.id == other.id

    # This is to ensure that the dict() function can accept this class.
    # See https://stackoverflow.com/a/40667249
    def keys(self):  # pylint: disable = missing-function-docstring
        return self._attrs.keys()

    def _deserialize_includes(self):
        for include, cls in self._includes.items():
            if include not in self:
                continue

            if isinstance(self._attrs[include], List):
                self._attrs[include] = [
                    cls(self.reporter, json_obj) for json_obj in self._attrs[include]
                ]
                if include in self._children:
                    getattr(  # pylint: disable = protected-access
                        self, include
                    )._list = self._attrs[include]
            else:
                self._attrs[include] = cls(self.reporter, self._attrs[include])


ChildOfRestObject = TypeVar("ChildOfRestObject", bound=RestObject)


class RestList(Sequence, Generic[ChildOfRestObject]):
    """Represents a list of :class:`~reporter.base.RestObject` instances built from server data.

    Includes associated links and metadata.

    Args:
        data: List of :class:`~reporter.base.RestObject` instances
        links: Dict of links (see Reporter API docs)
        meta: Dict of metadata (see Reporter API docs)
    """

    _data: List[ChildOfRestObject]
    links: Mapping[str, str]
    meta: Mapping[str, str]

    def __init__(
        self,
        data: List[ChildOfRestObject],
        links: Mapping[str, str],
        meta: Mapping[str, str],
    ) -> None:
        self._data = data
        self.links = links
        self.meta = meta

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)


class RestManager(Sequence, Generic[ChildOfRestObject]):
    """Base class for :class:`~reporter.base.RestObject` managers.

    Args:
        reporter: connection to use to make requests
        parent: :class:`~reporter.base.RestObject` to which the manager is attached, if applicable
    """

    reporter: Reporter

    _parent: Optional[RestObject]
    _parent_attrs: Mapping[str, str]
    _path: str
    _computed_path: str

    # We need to consider the case that an instance of this class is assigned
    # to an attribute of an instance `r` of RestObject with the same name as a
    # key in `r._includes`. This occurs e.g. for
    # `Assessment._children["targets"]`. Hence we make RestManager derive
    # collections.abc.Sequence and implement the required methods, so that this
    # class exposes a convenient API for the included list. We assume that this
    # can only happen if this class is assigned to an attribute of `r` that is
    # a plural noun (e.g. "targets" instead of "target").
    #
    # For example client.assessments refers to both the assessments manager and
    # a retrieved list of assessments, and can be used in both ways.
    _obj_cls: Type[ChildOfRestObject]
    _obj_cast: Callable
    _list: List[RestObject] = []

    def __init__(
        self,
        reporter: Reporter,
        parent: Optional[RestObject] = None,
    ) -> None:
        self.reporter = reporter
        self._parent = parent
        self._path = self._compute_path()

    def __getitem__(self, index):
        return self._list[index]

    def __len__(self):
        return len(self._list)

    def _compute_path(self) -> str:
        if self._parent is None:
            return self._path

        data = {
            pair[0]: getattr(self._parent, pair[1])
            for pair in self._parent_attrs.items()
        }
        return self._path.format(**data)



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import CrudMixin, ListMixin

__all__ = [
    "Client",
    "ClientManager",
]


class Client(RestObject):
    pass


class ClientManager(
    RestManager,
    CrudMixin,
    ListMixin,
):
    _path = "clients"
    _obj_cls = Client



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestManager, RestObject
from reporter.mixins import GetMixin, ListMixin

__all__ = [
    "AssessmentTemplate",
    "AssessmentTemplateManager",
]


class AssessmentTemplate(RestObject):
    pass


class AssessmentTemplateManager(RestManager, GetMixin, ListMixin):
    _path = "assessment-templates"
    _obj_cls = AssessmentTemplate



# pylint: disable = missing-module-docstring, missing-class-docstring

from reporter.base import RestObject

__all__ = [
    "FindingCreatedEvent",
]


class FindingCreatedEvent(RestObject):
    pass
